"""
Module listing all sankey related constants
"""
LABEL = 'label'
NODE_ID = 'node-id'
COLOR = 'color'
ROOT = 'root'
VALUE = 'value'
NUM_DATAPOINTS = 'num_datapoints'
NUM_CHILDREN = 'num-children'
ID = 'id'
DESCRIPTION = 'description'
HIERARCHIES_SHEET = 'Hierarchies'
HIERARCHY = 'hierarchy'
DEPTH = 'depth'